from typing import List, Tuple, Union
from .models import GenerationModel, PathModel
from .clip import clip
from PIL import Image
import torch

class T5QASummaryEmotion(GenerationModel):
    """Custom class for Kiri's T5 model for QA, Summarisation, Emotion tasks.

    \*args and \*\*kwargs according to GenerationModel class.
    
    model_path defaults to "kiri-ai/t5-base-qa-summary-emotion"
    """
    def __init__(self, *args, **kwargs):
        return super().__init__("kiri-ai/t5-base-qa-summary-emotion",
                                *args, **kwargs)


    def process_qa(self, question, context, prev_qa):
        input_text = [f"q: {qa[0]} a: {qa[1]}" for qa in prev_qa]
        input_text.append(f"q: {question}")
        input_text.append(f"c: {context}")
        input_text = " ".join(input_text)

        return input_text


    def qa(self, question, context, prev_qa: List[Tuple[str, str]] = []):
        self.check_init()
        if isinstance(question, list):
            # Must have a consistent amount of examples
            assert(len(question) == len(context))
            if len(prev_qa) != 0:
                assert(len(question) == len(prev_qa))
            else:
                prev_qa = [prev_qa] * len(question)

            # Process according to the model used
            input_text = [self.process_qa(q, c, p)
                          for q, c, p in zip(question, context, prev_qa)]
        else:
            input_text = self.process_qa(question, context, prev_qa)

        return self.generate(input_text, do_sample=False, max_length=96)

    
    def process_emotion(self, text):
        return f"emotion: {text}"

    
    def emotion(self, text):
        self.check_init()
        if isinstance(text, list):
            # Process according to the model used
            text = [self.process_emotion(item) for item in text]
        else:
            text = self.process_emotion(text)

        return self.generate(text, do_sample=False, max_length=96)

    
    def process_summarisation(self, text):
        return f"summarise: {text}"

    
    def summarise(self, text):
        self.check_init()
        if isinstance(text, list):
            # Process according to the model used
            text = [self.process_summarisation(item) for item in text]
        else:
            text = self.process_summarisation(text)

        return self.generate(text, do_sample=False, max_length=96)


class CLIP(PathModel):
    def __init__(self, model_path="ViT-B/32", init_model=clip.load, device=None, init=True):
        self.initialised = False
        self.init_model = init_model
        self.model_path = model_path
        self.device = device

        if self.device is None:
            self.device = "cuda" if torch.cuda.is_available() else "cpu"

        # Initialise
        if init:
            self.model, self.transform = self.init_model(model_path, device=self.device)
            
            self.initialised = True

    def __call__(self, image_path: str, labels: List[str]):
        return self.image_classification(image_path=image_path, labels=labels)        

    @torch.no_grad()
    def image_classification(self, image_path: Union[str, List[str]], labels: Union[List[str], List[List[str]]]):
        self.check_init()
        # TODO: Implement batching
        image = self.transform(Image.open(image_path)).unsqueeze(0).to(self.device)
        text = clip.tokenize(labels).to(self.device)

        image_features = self.model.encode_image(image)
        text_features = self.model.encode_text(text)
        
        logits_per_image, logits_per_text = self.model(image, text)
        probs = logits_per_image.softmax(dim=-1).cpu().numpy().tolist()[0]

        label_probs = zip(labels, probs)
        probabilities = {lp[0]: lp[1] for lp in label_probs}
        return probabilities


from typing import Callable, Dict, List, Tuple

from .search import DocStore, SearchResults, Document, InMemoryDocStore, ChunkedDocument
from .utils import process_documents, process_results
from .models import Generation, Vectorisation, Summarisation, \
    Emotion, QA, Classification, T5QASummaryEmotion, ImageClassification

import logging


class Kiri:
    """Core class of natural language engine

    Attributes:
        store (optional): DocStore object to be used as the engine backend
        vectorisation_model (optional): "english" or "multilingual".
            For local inference, the name of a SentenceTransformer model is also supported.
        classification_model (optional): "english" or "multilingual".
            For local inference, the name of a Huggingface transformers model is also supported.
        process_doc_func (optional): Function to be used when vectorising uploaded documents
        process_results_func (optional): Function to be used for calculating final scores of results
        device (optional): Pytorch device to run inference on. Detected automatically if not specified.
    """

    def __init__(self, store: DocStore = None, local=False, api_key=None,
                 vectorisation_model: str = None,
                 classification_model: str = None,
                 image_classification_model: str = None,
                 generation_model: str = None,
                 process_doc_func: Callable[[
                     Document, str], List[float]] = None,
                 process_results_func: Callable[[
                     SearchResults, str], None] = None,
                 device: str = None):

        if store is None:
            store = InMemoryDocStore()

        store.kiri = self

        if local and not device:
            import torch
            device = "cuda" if torch.cuda.is_available() else "cpu"

        if process_doc_func is None:
            # Use default vectoriser
            process_doc_func = process_documents

        if process_results_func is None:
            process_results_func = process_results

        t5_qa_summary_emotion = T5QASummaryEmotion(device=device, init=False)

        self._vectorise = Vectorisation(vectorisation_model, local=local,
                                        api_key=api_key, device=device, init=False)
        
        self._generate = Generation(generation_model, local=local,
                                        api_key=api_key, device=device, init=False)

        self._classify = Classification(classification_model, local=local,
                            api_key=api_key, device=device, init=False)

        self._image_classification = ImageClassification(image_classification_model, local=local,
                            api_key=api_key, device=device, init=False)

        self._qa = QA(t5_qa_summary_emotion, local=local,
                            api_key=api_key, device=device, init=False)
        
        self._emotion = Emotion(t5_qa_summary_emotion, local=local,
                            api_key=api_key, device=device, init=False)

        self._summarise = Summarisation(t5_qa_summary_emotion, local=local,
                            api_key=api_key, device=device, init=False)
        
        self._device = device
        self._local = local
        self._api_key = api_key
        self._store = store
        self._process_doc_func = process_doc_func
        self._process_results_func = process_results

    def upload(self, documents: List[Document]) -> None:
        """Processes and uploads documents to store

        Args:
            documents: List of documents for upload

        Returns:
            None

        Example:
            >>> kiri.upload([Document("First document"), Document("Second document")])
            None

        """
        logging.warning("Upload functionality is deprecated and will be removed in a future version. Use https://github.com/kiri-ai/kiri-search instead.")
        return self._store.upload(documents, self._process_doc_func)

    def search(self, query: str, max_results=10, min_score=0.0,
               preview_length=100, ids: List[str] = [], body=None) -> SearchResults:
        """Search documents from document store

        Args:
            query: Search string on which search is performed
            max_results: Maximum amount of documents to be returned from search
            min_score: Minimum score required to be included in results
            preview_length: Number of characters in the preview/metatext of results
            ids: List of ids to search from
            body: Elasticsearch request body to be passed to the backend

        Returns:
            SearchResults object

        Example:
            >>> kiri.search("RTX 3090")
            SearchResults object

        """
        logging.warning("Search functionality is deprecated and will be removed in a future version. Use https://github.com/kiri-ai/kiri-search instead.")
        search_results, query_vec = self._store.search(query,
                                                       max_results=max_results, min_score=min_score,
                                                       ids=ids, body=body)
        self._process_results_func(
            search_results, query_vec, self._store._doc_class,
            preview_length, max_results, min_score)
        return search_results

    def qa(self, question: str, context: str = None,
           prev_qa: List[Tuple[str, str]] = [], num_answers: int = 3):
        """Perform QA, either on docstore or on provided context.

        Args:
            question: Question (string or list of strings if using own context) for qa model.
            context (optional): Context (string or list of strings) to ask question from.
            prev_qa (optional): List of previous question, answer tuples or list of prev_qa.
            num_answers (optional): Number of answers to return

        Returns:
            if context is given: Answer string or list of answer strings
            if no context: List of num_answers (default 3) answer and SearchResult object pairs.

        Example:
            >>> kiri.qa("Where does Sally live?", "Sally lives in London.")
            "London"
        """
        if context:
            return self._qa(question, context, prev_qa=prev_qa)
        else:
            logging.warning("No context (search based) qa is deprecated and will be removed in a future version. Use https://github.com/kiri-ai/kiri-search instead.")
            search_results = self.search(question)
            answers = []
            if issubclass(self._store._doc_class, ChunkedDocument):
                for chunk in search_results.top_chunks[:num_answers]:
                    answer = self._qa(question, chunk["chunk"], prev_qa=prev_qa)
                    answers.append((answer, chunk["search_result"]))
            else:
                for result in search_results.results[:num_answers]:
                    c_string = result.document.content
                    answer = self._qa(question, c_string, prev_qa=prev_qa)
                    answers.append((answer, result))
            return answers


    def summarise(self, input_text):
        """Perform summarisation on input text.

        Args:
            input_text: string or list of strings to be summarised - keep each string below 500 words.

        Returns:
            Summary string or list of summary strings.

        Example:
            >>> kiri.summarise("This is a long document that contains plenty of words")
            "short summary of document"

        """
        return self._summarise(input_text)

    def emotion(self, input_text):
        """Perform emotion detection on input text.

        Args:
            input_text: string or list of strings to detect emotion from
                keep this under a few sentences for best performance.

        Returns:
            Emotion string or list of emotion strings.
                Each emotion string contains comma and space separated emotions.
            Emotions are from: admiration, approval, annoyance, gratitude, disapproval, amusement,
                curiosity, love, optimism, disappointment, joy, realization, anger, sadness, confusion,
                caring, excitement, surprise, disgust, desire, fear, remorse, embarrassment, nervousness,
                pride, relief, grief

        Example:
            >>> kiri.emotion("I really like what you did there")
            "approval"

        """
        return self._emotion(input_text)

    def classify(self, input_text, labels: List[str]):
        """Classify input text according to given labels.


        Args:
            input_text: string or list of strings to classify
            labels: list of strings or list of labels

        Returns:
            dict where each key is a label and value is probability between 0 and 1, or list of dicts.

        Example:
            >>> kiri.classify("I am mad because my product broke.", ["product issue", "nature"])
            {"product issue": 0.98, "nature": 0.05}

        """
        return self._classify(input_text, labels)


    def image_classification(self, image_path: str, labels: List[str]):
        # TODO: Implement batching
        """Classify image according to given labels.


        Args:
            image_path: path to image
            labels: list of strings

        Returns:
            dict where each key is a label and value is probability between 0 and 1

        Example:
            >>> kiri.image_classification("/home/Documents/dog.png", ["cat", "dog"])
            {"cat": 0.01, "dog": 0.99}

        """
        return self._image_classification(image_path, labels)


    def vectorise(self, input_text):
        """Vectorise input text.


        Args:
            input_text: string or list of strings to vectorise

        Returns:
            Vector or list of vectors

        Example:
            >>> kiri.vectorise("iPhone 12 128GB")
            [0.92949192, 0.23123010, ...]

        """
        return self._vectorise(input_text)

    def generate(self, input_text, min_length=10, max_length=20, temperature=1.0,
                top_k=0.0, top_p=1.0, repetition_penalty=1.0, length_penalty=1.0,
                num_beams=1, num_generations=1, do_sample=True):
        """Generates text to continue off the given input.

        Args:
            input_text: Text from which model will begin generating.
            min_length: Minimum length of generation before EOS can be generated.
            max_length: Maximum length of generated sequence.
            temperature: Value that alters softmax probabilities.
            top_k: Sampling strategy in which probabilities are redistributed among top k most-likely words.
            top_p: Sampling strategy in which probabilities are distributed among 
                set of words with combined probability greater than p.
            repetition_penalty: Penalty to be applied to words present in the input_text and
                words already generated in the sequence.
            length_penalty: Penalty applied to overall sequence length. Set >1 for longer sequences,
                or <1 for shorter ones. 
            num_beams: Number of beams to be used in beam search. (1: no beam search)
            num_generations: How many times to run generation. 
            do_sample: Whether or not sampling strategies (top_k & top_p) should be used.
        """
        return self._generate(input_text,
                          min_length=min_length,
                          max_length=max_length, temperature=temperature,
                          top_k=top_k, top_p=top_p, repetition_penalty=repetition_penalty,
                          length_penalty=length_penalty, num_beams=num_beams,
                          num_return_sequences=num_generations, do_sample=do_sample)


from typing import List, Tuple, Union
from ..models import BaseModel, ClassificationModel
from .base import Task

import requests

DEFAULT_LOCAL_MODEL = "facebook/bart-large-mnli"

LOCAL_MODELS = {
    "english": DEFAULT_LOCAL_MODEL,
    "multilingual": "joeddav/xlm-roberta-large-xnli"
}

DEFAULT_API_MODEL = "english"

API_MODELS = ["english", "multilingual"]

class Classification(Task):
    """
    Task for classification.

    Attributes:
        model:
            1. Name of the model on Kiri's classification endpoint (english, multilingual)
            2. Officially supported local models (english, multilingual) or Huggingface path to the model.
            3. Kiri's ClassificationModel object
        local (optional): Run locally. Defaults to True
        api_key (optional): Kiri API key for non-local inference
        device (optional): Device to run inference on. Defaults to "cuda" if available.
        init (optional): Whether to initialise model immediately or wait until first call.
            Defaults to False
    """
    def __init__(self, model: Union[str, BaseModel] = None,
                local: bool = False, api_key: str = None, device: str = "cpu",
                init: bool = False):

        super().__init__(model, local=local, api_key=api_key, device=device,
                        init=init, local_models=LOCAL_MODELS, api_models=API_MODELS,
                        default_local_model=DEFAULT_LOCAL_MODEL,
                        default_api_model=DEFAULT_API_MODEL)
        
        # Model must implement the task
        if self.local:
            # Still needs to be initialised
            if type(self.model) == str:
                self.model = ClassificationModel(self.model, init=init, device=device)

            task = getattr(self.model, "classify", None)
            if not callable(task):
                raise ValueError(f"The model {model} cannot be used for classification.\
                                It does not implement the 'classify' method.")
    
    def __call__(self, text: str, labels: List[str]):
        """
        Calls the classification model with text and labels.
        """
        if self.local:
            return self.model.classify(text, labels)
        else:
            body = {
                "text": text,
                "labels": labels,
                "model": self.model,
            }

            res = requests.post("https://api.kiri.ai/classification", json=body,
                                headers={"x-api-key": self.api_key}).json()

            return res["probabilities"]


from .tasks.qa import QA
from .tasks.summarisation import Summarisation
from .tasks.emotion import Emotion
from .tasks.classification import Classification
from .tasks.image_classification import ImageClassification
from .tasks.vectorisation import Vectorisation
from .tasks.generation import Generation
from .custom_models import T5QASummaryEmotion, CLIP
from .models import BaseModel, PathModel, HuggingModel, GenerationModel, ClassificationModel



from transformers import AutoModelForPreTraining, AutoTokenizer, \
    AutoModelForSequenceClassification
from sentence_transformers import SentenceTransformer
import torch

class BaseModel:
    """
    The base class for a model.
    
    Attributes:
        model: Your model that takes some args, kwargs and returns an output.
            Must be callable.
    """
    def __init__(self, model):
        self.model = model

    def __call__(self, *args, **kwargs):
        return self.model(*args, **kwargs)


class PathModel(BaseModel):
    """
    Class for models which are initialised from a path.

    Attributes:
        model_path: Path to the model
        init_model: Callable to initialise model from path
        tokenizer_path (optional): Path to the tokenizer
        init_tokenizer (optional): Callable to initialise tokenizer from path
        device (optional): Device for inference. Defaults to "cuda" if available.
        init (optional): Whether to initialise model and tokenizer immediately or wait until first call.
            Defaults to True.
    """
    def __init__(self, model_path, init_model, tokenizer_path=None,
                init_tokenizer=None, device=None, init=True):
        self.initialised = False
        self.init_model = init_model
        self.init_tokenizer = init_tokenizer
        self.model_path = model_path
        self.tokenizer_path = tokenizer_path
        self.device = device

        if self.device is None:
            self.device = "cuda" if torch.cuda.is_available() else "cpu"

        # Initialise
        if init:
            self.model = self.init_model(model_path).eval().to(self.device)

            # Not all models need tokenizers
            if self.tokenizer_path:
                self.tokenizer = self.init_tokenizer(self.tokenizer_path)
            
            self.initialised = True


    def check_init(self):
        # Initialise on call time
        if not self.initialised:
            self.__init__(self.model_path, init=True)


    def __call__(self, *args, **kwargs):
        self.check_init()

        return self.model(*args, **kwargs)


class HuggingModel(PathModel):
    """
    Class for models which are initialised from a local path or huggingface

    Attributes:
        model_path: Local or huggingface.co path to the model
        init_model: Callable to initialise model from path
            Defaults to AutoModelForPreTraining from huggingface
        tokenizer_path (optional): Path to the tokenizer
        init_tokenizer (optional): Callable to initialise tokenizer from path
            Defaults to AutoTokenizer from huggingface.
        device (optional): Device for inference. Defaults to "cuda" if available.
        init (optional): Whether to initialise model and tokenizer immediately or wait until first call.
            Defaults to True.
    """
    def __init__(self, model_path, tokenizer_path=None,
                model_class=AutoModelForPreTraining,
                tokenizer_class=AutoTokenizer, device=None, init=True):
        # Usually the same
        if not tokenizer_path:
            tokenizer_path = model_path

        # Object was made with init = False
        if hasattr(self, "initialised"):
            model_path = self.model_path
            tokenizer_path = self.tokenizer_path
            init_model = self.init_model
            init_tokenizer = self.init_tokenizer
            device = self.device
        else:
            init_model = model_class.from_pretrained
            init_tokenizer = tokenizer_class.from_pretrained

        return super().__init__(model_path, tokenizer_path=tokenizer_path,
                                init_model=init_model,
                                init_tokenizer=init_tokenizer,
                                device=device, init=init)


class VectorisationModel(PathModel):
    """
    Class for models which are initialised from a local path or Sentence Transformers

    Attributes:
        model_path: Local, sentence transformers or huggingface.co path to the model
        init_model: Callable to initialise model from path
            Defaults to SentenceTransformer
        device (optional): Device for inference. Defaults to "cuda" if available.
        init (optional): Whether to initialise model immediately or wait until first call.
            Defaults to True.
    """
    def __init__(self, model_path, model_class=SentenceTransformer,
                device=None, init=True):
        # Object was made with init = False
        if hasattr(self, "initialised"):
            model_path = self.model_path
            init_model = self.init_model
            device = self.device
        else:
            init_model = model_class

        return super().__init__(model_path,
                                init_model=init_model,
                                device=device, init=init)


    def __call__(self, *args, **kwargs):
        return self.vectorise(*args, **kwargs)

    def vectorise(self, *args, **kwargs):
        self.check_init()
        with torch.no_grad():
            return self.model.encode(*args, **kwargs)


class GenerationModel(HuggingModel):
    """
    Class for models which are initialised from a local path or Sentence Transformers

    Attributes:
        *args and **kwargs are passed to HuggingModel's __init__
    """
    def generate(self, text, **kwargs):
        """
        Generate according to the model's generate method.
        """
        self.check_init()

        # Get and remove do_sample or set to False
        do_sample = kwargs.pop("do_sample", None) or False
        params = ["temperature", "top_k", "top_p", "repetition_penalty",
                    "length_penalty", "num_beams", "num_return_sequences"]

        # If params are changed, we want to sample
        for param in params:
            if param in kwargs.keys():
                do_sample = True
                break

        is_list = False
        if isinstance(text, list):
            is_list = True

        if not is_list:
            text = [text]

        all_tokens = []
        for text in text:
            features = self.tokenizer(text, return_tensors="pt")

            for k, v in features.items():
                features[k] = v.to(self.device)

            with torch.no_grad():
                tokens = self.model.generate(do_sample=do_sample,
                                            **features, **kwargs)

            all_tokens.append(tokens)
            
        value = []
        for tokens in all_tokens:
            value.append([self.tokenizer.decode(tokens, skip_special_tokens=True)
                    for tokens in tokens])
        
        output = value

        # Unwrap generation list
        if kwargs.get("num_return_sequences", 1) == 1:
            output_unwrapped = []
            for value in output:
                output_unwrapped.append(value[0])

            output = output_unwrapped
        
        # Return single item
        if not is_list:
            output = output[0]

        return output


class ClassificationModel(HuggingModel):
    """
    Class for classification models which are initialised from a local path or huggingface

    Attributes:
        model_path: Local or huggingface.co path to the model
        tokenizer_path (optional): Path to the tokenizer
        model_class (optional): Callable to initialise model from path
            Defaults to AutoModelForSequenceClassification from huggingface
        tokenizer_class (optional): Callable to initialise tokenizer from path
            Defaults to AutoTokenizer from huggingface.
        device (optional): Device for inference. Defaults to "cuda" if available.
        init (optional): Whether to initialise model and tokenizer immediately or wait until first call.
            Defaults to True.
    """
    def __init__(self, model_path, tokenizer_path=None,
                model_class=AutoModelForSequenceClassification,
                tokenizer_class=AutoTokenizer, device=None, init=True):
        return super().__init__(model_path, tokenizer_path=tokenizer_path,
                    model_class=model_class, tokenizer_class=tokenizer_class,
                    device=device, init=init)

    def calculate_probability(self, text, label, device):
        hypothesis = f"This example is {label}."
        features = self.tokenizer.encode(text, hypothesis, return_tensors="pt",
                                    truncation=True).to(self.device)
        logits = self.model(features)[0]
        entail_contradiction_logits = logits[:, [0, 2]]
        probs = entail_contradiction_logits.softmax(dim=1)
        prob_label_is_true = probs[:, 1]
        return prob_label_is_true.item()


    def classify(self, text, labels):
        """
        Classifies text, given a set of labels.
        """
        self.check_init()
        if isinstance(text, list):
            # Must have a consistent amount of examples
            assert(len(text) == len(labels))
            # TODO: implement proper batching
            results_list = []
            for text, labels in zip(text, labels):
                results = {}
                for label in labels:
                    results[label] = self.calculate_probability(text, label, self.device)

                results_list.append(results)

            return results_list
        else:
            results = {}
            for label in labels:
                results[label] = self.calculate_probability(
                    text, label, self.device)

            return results


import pytest
from kiri import Kiri, Document, ChunkedDocument

text = "Jon likes beer. He drinks it at pubs every Friday."


def test_make_doc():
    c = "I am a document."
    doc = Document(c)
    assert doc.content == c


def test_make_chunk_doc():
    c = "I am a document. This is a sentence."
    doc = ChunkedDocument(c, chunking_level=1)
    assert doc.content == c
    assert doc.chunking_level == 1


def test_qa():
    kiri = Kiri(local=True)
    out = kiri.qa("What does Jon like?", text)
    assert type(out) == str


def test_summarise():
    kiri = Kiri(local=True)
    out = kiri.summarise(text)
    assert type(out) == str


def test_emotion():
    kiri = Kiri(local=True)
    out = kiri.emotion(text)
    assert type(out) == str


def test_classify():
    kiri = Kiri(local=True)
    out = kiri.classify(text, ["interests", "alcoholism"])
    assert isinstance(out, dict)


def test_doc_qa():
    kiri = Kiri(local=True)
    c = "I am a document."
    doc = Document(c)
    kiri.upload([doc])
    out = doc.qa("What are you?")
    assert type(out) == str


def test_doc_summarise():
    kiri = Kiri(local=True)
    c = "I am a document."
    doc = Document(c)
    kiri.upload([doc])
    out = doc.summarise()
    assert type(out) == str


def test_doc_summarise():
    kiri = Kiri(local=True)
    c = "I am a document."
    doc = Document(c)
    kiri.upload([doc])
    out = doc.emotion()
    assert type(out) == str


def test_doc_summarise():
    kiri = Kiri(local=True)
    c = "I am a document."
    doc = Document(c)
    kiri.upload([doc])
    out = doc.classify(["random"])
    assert isinstance(out, dict)



from typing import List, Tuple, Union
from ..models import BaseModel, GenerationModel
from .base import Task

import requests

DEFAULT_LOCAL_MODEL = "gpt2"

LOCAL_MODELS = {
    "gpt2": "gpt2",
    "t5-base-qa-summary-emotion": "t5-base-qa-summary-emotion"
}

DEFAULT_API_MODEL = "gpt2-large"

API_MODELS = ["gpt2-large", "t5-base-qa-summary-emotion"]

class Generation(Task):
    """
    Task for text generation.

    Attributes:
        model:
            1. Name of the model on Kiri's generation endpoint (gpt2-large, t5-base-qa-summary-emotion)
            2. Officially supported local models (gpt2, t5-base-qa-summary-emotion) or Huggingface path to the model.
            3. Kiri's GenerationModel object
        local (optional): Run locally. Defaults to True
        api_key (optional): Kiri API key for non-local inference
        device (optional): Device to run inference on. Defaults to "cuda" if available.
        init (optional): Whether to initialise model immediately or wait until first call.
            Defaults to False
    """
    def __init__(self, model: Union[str, BaseModel] = None,
                local: bool = False, api_key: str = None, device: str = "cpu",
                init: bool = False):

        super().__init__(model, local=local, api_key=api_key, device=device,
                        init=init, local_models=LOCAL_MODELS, api_models=API_MODELS,
                        default_local_model=DEFAULT_LOCAL_MODEL,
                        default_api_model=DEFAULT_API_MODEL)
        
        # Model must implement the task
        if self.local:
            # Still needs to be initialised
            if type(self.model) == str:
                self.model = GenerationModel(self.model, init=init, device=device)

            task = getattr(self.model, "generate", None)
            if not callable(task):
                raise ValueError(f"The model {model} cannot be used for generation.\
                                It does not implement the 'generate' method.")
    
    def __call__(self, text: str, *args, **kwargs):
        """
        Calls the generation model with text and generation *args, **kwargs.
        """
        if self.local:
            return self.model.generate(text, *args, **kwargs)
        else:
            body = {
                "text": text,
                "model": self.model,
                **kwargs
            }

            res = requests.post("https://api.kiri.ai/generation", json=body,
                                headers={"x-api-key": self.api_key}).json()

            return res["output"]


from kiri import Kiri, Document, ChunkedDocument
from kiri.search import SearchResult
import pytest


def get_docs():
    doc1 = Document("Hello I am a document. This is a sentence")
    doc2 = Document("Hello I am another document. This is another sentence")
    docs = [doc1, doc2]
    return docs


def get_chunked_docs(chunking_level=1):
    doc1 = ChunkedDocument(
        "Hello I am a document. This is a sentence", chunking_level=chunking_level)
    doc2 = ChunkedDocument(
        "Hello I am another document. This is another sentence", chunking_level=chunking_level)
    docs = [doc1, doc2]
    return docs


def test_init():
    kiri = Kiri(local=True)


def test_upload():
    kiri = Kiri(local=True)
    docs = get_docs()
    kiri.upload(docs)
    assert docs[0].vector is not None, "Document not vectorised"
    assert docs[1].vector is not None, "Document not vectorised"
    assert len(kiri._store.documents) == 2, "Incorrect number of documents in mem"


def test_upload_chunked():
    kiri = Kiri(local=True)
    docs = get_chunked_docs(chunking_level=1)
    kiri.upload(docs)
    assert len(kiri._store.documents) == 2, "Incorrect number of documents in mem"
    for doc in docs:
        assert doc.vector is not None, "Document not vectorised"
        assert len(doc.chunk_vectors) == 2, "Invalid number of chunk vectors"


def test_upload_dup_id():
    kiri = Kiri(local=True)
    docs = get_docs()
    for doc in docs:
        doc.id = "123"

    with pytest.raises(ValueError):
        kiri.upload(docs)


def test_upload_mixed_type():
    kiri = Kiri(local=True)
    docs = [Document("a"), ChunkedDocument("b")]

    with pytest.raises(ValueError):
        kiri.upload(docs)


def test_search():
    kiri = Kiri(local=True)
    docs = get_docs()
    kiri.upload(docs)
    results = kiri.search("another")
    assert len(results.results) == 2, "Invalid number of search results"


def test_search_max_results():
    kiri = Kiri(local=True)
    docs = get_docs()
    kiri.upload(docs)
    results = kiri.search("another", max_results=1)
    assert len(results.results) == 1, "Invalid number of search results"


def test_search_ids():
    kiri = Kiri(local=True)
    docs = get_docs()
    docs[0].id = "123"
    kiri.upload(docs)
    results = kiri.search("another", ids=["123"])
    assert len(results.results) == 1, "Invalid number of search results"


def test_search_chunk():
    kiri = Kiri(local=True)
    docs = get_chunked_docs()
    kiri.upload(docs)
    results = kiri.search("another")
    assert len(results.results) == 2, "Invalid number of search results"


def test_search_max_results_chunk():
    kiri = Kiri(local=True)
    docs = get_chunked_docs()
    kiri.upload(docs)
    results = kiri.search("another", max_results=1)
    assert len(results.results) == 1, "Invalid number of search results"


def test_search_ids_chunk():
    kiri = Kiri(local=True)
    docs = get_chunked_docs()
    docs[0].id = "123"
    kiri.upload(docs)
    results = kiri.search("another", ids=["123"])
    assert len(results.results) == 1, "Invalid number of search results"


def test_qa():
    kiri = Kiri(local=True)
    docs = get_docs()
    kiri.upload(docs)
    results = kiri.qa("another?")
    assert isinstance(results, list)
    for result in results:
        assert type(result[0]) == str
        assert isinstance(result[1], SearchResult)


def test_qa_chunk():
    kiri = Kiri(local=True)
    docs = get_chunked_docs()
    kiri.upload(docs)
    results = kiri.qa("another?")
    assert isinstance(results, list)
    for result in results:
        assert type(result[0]) == str
        assert isinstance(result[1], SearchResult)



from typing import List, Tuple, Union
from ..models import BaseModel
from ..custom_models import T5QASummaryEmotion
from .base import Task

import requests

DEFAULT_LOCAL_MODEL = "kiri-ai/t5-base-qa-summary-emotion"

LOCAL_MODELS = {
    "english": DEFAULT_LOCAL_MODEL
}

DEFAULT_API_MODEL = "english"

API_MODELS = ["english"]

class Emotion(Task):
    """
    Task for emotion detection.

    Attributes:
        model:
            1. Name of the model on Kiri's emotion endpoint (english)
            2. Officially supported local models (english) or Huggingface path to the model.
            3. Kiri's BaseModel object that implements the emotion method
        model_class (optional): The model class to use when supplying a path for the model.
        local (optional): Run locally. Defaults to True
        api_key (optional): Kiri API key for non-local inference
        device (optional): Device to run inference on. Defaults to "cuda" if available.
        init (optional): Whether to initialise model immediately or wait until first call.
            Defaults to False
    """
    def __init__(self, model: Union[str, BaseModel] = None, model_class=T5QASummaryEmotion,
                local: bool = False, api_key: str = None, device: str = "cpu",
                init: bool = False):

        super().__init__(model, local=local, api_key=api_key, device=device,
                        init=init, local_models=LOCAL_MODELS, api_models=API_MODELS,
                        default_local_model=DEFAULT_LOCAL_MODEL,
                        default_api_model=DEFAULT_API_MODEL)
        
        # Model must implement the task
        if self.local:
            # Still needs to be initialised
            if type(self.model) == str:
                self.model = model_class(self.model, init=init, device=device)

            task = getattr(self.model, "emotion", None)
            if not callable(task):
                raise ValueError(f"The model {model} cannot be used for emotion.\
                                It does not implement the 'emotion' method.")
    
    def __call__(self, text: str):
        """
        Calls the emotion model with text.
        """
        if self.local:
            return self.model.emotion(text)
        else:
            body = {
                "text": text,
                "model": self.model
            }

        res = requests.post("https://api.kiri.ai/emotion", json=body,
                            headers={"x-api-key": self.api_key}).json()

        return res["emotion"]


from typing import Dict, List
import logging

logger = logging.getLogger("info")

class Task:
    def __init__(self, model, local=False, api_key=None,
                device: str = "cpu", init=False, local_models: Dict = None,
                api_models: List[str] = None, default_local_model: str = None,
                default_api_model: str = None):
        self.local = local
        self.api_key = api_key
        self.device = device
        self.init = init

        # Pick the correct model name
        if local:
            if model is None:
                model = default_local_model

            if type(model) == str or model is None:
                # Get from dictionary or use provided if not there
                model = local_models.get(model) or model
                if model not in local_models.values():
                    logger.warning(f"Model '{model}' is not officially supported, but it may function.")
                
            self.model = model
        else:
            if model is None or type(model) != str:
                model = default_api_model

            if api_key is None:
                raise ValueError(
                    "Please provide your api_key (https://kiri.ai) with api_key=... or set local=True")
            
            if model not in api_models:
                raise ValueError(f"Model '{model}' is not supported for this task. Please use one of: {', '.join(api_models)}")
    
            # All checks passed
            self.model = model

    def __call__(self):
        raise Exception("The base Task is not callable!")


# TODO: Rename test_models.py to test_tasks.py and name imports better
from kiri.models import Vectorisation, QA, Summarisation, Emotion, Classification, ImageClassification
import numpy as np
import torch
import os

qa_example = {
    "q1": "Where does Sally live?",
    "q2": "How long has Sally lived in London?",
    "q2_conv": "How long has she lived there?",
    "q3": "Where did Sally live prior to London?",
    "q3_conv": "Where did she live before?",
    "c": "Sally has been living in London for 3 years. Previously, Sally lived in Liverpool.",
    "a1": "London",
    "a2": "3 years",
    "a3": "Liverpool"
}

summary_context = """Elon Reeve Musk FRS (/ˈiːlɒn/ EE-lon; born June 28, 1971) is a business magnate, industrial designer and engineer.[6] He is the founder, CEO, CTO and chief designer of SpaceX; early investor,[b] CEO and product architect of Tesla, Inc.; founder of The Boring Company; co-founder of Neuralink; and co-founder and initial co-chairman of OpenAI. He was elected a Fellow of the Royal Society (FRS) in 2018.[9][10] Also that year, he was ranked 25th on the Forbes list of The World's Most Powerful People,[11] and was ranked joint-first on the Forbes list of the Most Innovative Leaders of 2019.[12] As of December 19, 2020, Musk’s net worth was estimated by Forbes to US$153.5 billion,[1][13] making him the second-richest person in the world, behind Jeff Bezos.[14]"""
classify_context = """I am mad because my product broke the first time I used it"""
classify_labels = ["product issue", "nature"]
classify_correct = "product issue"

main_path = os.path.dirname(__file__)
image_classification_image = os.path.join(main_path, "data/dog.png")
image_classification_labels = ["cat", "dog"]
image_classification_correct = "dog"

device = "cuda" if torch.cuda.is_available() else "cpu"

vectorise = Vectorisation(local=True, device=device)
qa = QA(local=True, device=device)
emotion = Emotion(local=True, device=device)
summarise = Summarisation(local=True, device=device)
classify = Classification(local=True, device=device)
image_classification = ImageClassification(local=True, device=device)

def test_vectorisation_single():
    out = vectorise("This is a sample thing.")
    assert isinstance(out, np.ndarray), "Not an array"
    assert type(out[0]) == np.dtype("float32"), "Vector dimension not float32"


def test_vectorisation_batch():
    out = vectorise(["This is a sample thing.",
                     "This is another sample thing."])
    assert isinstance(out, np.ndarray), "Not an array"
    assert len(out) == 2, "Not the right size"
    assert type(out[0][0]) == np.dtype(
        "float32"), "Vector dimension not float32"
    assert not np.array_equal(
        out[0], out[1]), "Two different texts have same vector"


def test_qa_single():
    out = qa(qa_example["q1"], qa_example["c"])
    assert type(out) == str, "Qa answer not a string"
    assert out == qa_example["a1"], "Wrong answer to simple qa"


def test_qa_batch():
    out = qa([qa_example["q1"], qa_example["q2"]],
             [qa_example["c"], qa_example["c"]])
    assert isinstance(out, list), "Output is not a list"
    assert len(out) == 2, "Incorrect number of answers"
    assert out[0] == qa_example["a1"], "Wrong answer to simple qa"
    assert out[1] == qa_example["a2"], "Wrong answer to simple qa"


def test_qa_conv_single():
    out = qa(qa_example["q2_conv"], qa_example["c"], prev_qa=[
             (qa_example["q1"], qa_example["a1"])])
    assert type(out) == str, "Qa answer not a string"
    assert out == qa_example["a2"], "Wrong answer to simple conversational qa"


def test_qa_conv_batch():
    questions = [qa_example["q2_conv"], qa_example["q3_conv"]]
    ctxs = [qa_example["c"], qa_example["c"]]
    prev_qa = [[(qa_example["q1"], qa_example["a1"])], [
        (qa_example["q2"], qa_example["a2"])]]

    out = qa(questions, ctxs, prev_qa=prev_qa)
    assert isinstance(out, list), "Output not a list"
    assert len(out) == 2, "Incorrect number of answers"
    assert out[0] == qa_example["a2"], "Wrong answer to simple conversational qa"
    assert out[1] == qa_example["a3"], "Wrong answer to simple conversational qa"


def test_summary_single():
    out = summarise(summary_context)
    assert type(out) == str, "Summary not a string"
    assert len(out) < len(summary_context), "Summary not shorter than input"


def test_summary_bulk():
    out = summarise([summary_context, summary_context])
    assert isinstance(out, list), "Output not a list"
    assert len(out) == 2, "Incorrect number of outputs"
    assert len(out[0]) < len(summary_context), "Summary not shorter than input"
    assert len(out[1]) < len(summary_context), "Summary not shorter than input"


def test_classify_single():
    out = classify(classify_context, classify_labels)
    assert isinstance(out, dict), "Output is not a dict"
    assert len(out.keys()) == len(
        classify_labels), "Incorrect number of labels"
    assert classify_correct == max(
        out, key=out.get), "Classification is severely wrong"


def test_classify_batch():
    out = classify([classify_context, classify_context],
                    [classify_labels, classify_labels])
    assert isinstance(out, list), "Output is not a list"

    for out in out:
        assert isinstance(out, dict), "List item is not a dict"
        assert len(out.keys()) == len(
            classify_labels), "Incorrect number of labels"
        assert classify_correct == max(
            out, key=out.get), "Classification is severely wrong"


def test_emotion_single():
    out = emotion("I am really angry.")
    assert type(out) == str, "Output is not a string"


def test_emotion_batch():
    out = emotion(["I am really angry.", "You are really angry."])
    assert isinstance(out, list), "Output is not a list"

    for out in out:
        assert type(out) == str, "List not made of strings"


def test_image_classification_single():
    out = image_classification(image_classification_image, image_classification_labels)
    assert isinstance(out, dict), "Output is not a dict"
    assert len(out.keys()) == len(
        image_classification_labels), "Incorrect number of labels"
    assert image_classification_correct == max(
        out, key=out.get), "Classification is severely wrong"


def test_image_classification_batch():
    out = image_classification([image_classification_image]*2, [image_classification_labels]*2)

    assert isinstance(out, list), "Output is not a list"

    for out in out:
        assert len(out.keys()) == len(
            image_classification_labels), "Incorrect number of labels"
        assert image_classification_correct == max(
            out, key=out.get), "Classification is severely wrong"
